#!/usr/bin/python
import math
import random
import os
import time
#from termcolor import colored

'''
       Dieses Programm realisiert ein einfaches Raeuber-Beute-Modell auf der
       Basis von zellulaeren Automaten.
       Es gibt Hasen (H) und Fuechse (F), die sich vermehren koennen.
       Hasen koennen von Fuechsen gefressen werden, Fuechse sterben, wenn
       sie keine Hasen finden.
       Eingabeparameter:
            n:    Anzahl von Zellen in jeder Dimension (z.B. 40)
            ngen: Anzahl von Generationen, die betrachtet werden sollen

'''

def clearmat(a,n):
   #Setze Matrix a auf 0
   for i in range(n+2):
      for j in range(n+2):
         a[i][j]=0

def copymat(a,b,n):
   #Setze Matrix a auf 0
   for i in range(n+2):
      for j in range(n+2):
         b[i][j]=a[i][j]

def analyze_population(a,n,history,ngen,it):
   # Analysiere aktuelle Populationen im Schritt it und speichere diese Infos
   # in history in der Reihenfolge "Generation, Anzahl der Hasen, Anzahl der Fuechse"
   history[0][it]=it
   history[1][it]=0
   history[2][it]=0
   for i in range(1,n+1):
      for j in range(1,n+1):
         if a[i][j]==-10:
            history[2][it]=history[2][it]+1
         elif a[i][j]==1:
            history[1][it]=history[1][it]+1

def print_history(history,ngenp):
   # Drucke Historie der Populationen aus
   print
   print "Generation   Hasen    Fuechse"
   print "-----------------------------"
   for i in range(ngenp):
      print "%6i%11i%10i" %(history[0][i], history[1][i], history[2][i])


def printmat(a,n):
   # Drucke Inneres der Matrix a; Fuechse ("-10") werden als "F" und Hasen ("1")
   # werden als "H" ausgegeben
   for i in range(1,n+1):
      for j in range(1,n+1):
         if a[i][j]==-10:
            #print colored("F","red"),
            print "F",
         elif a[i][j]==1:
            print "H",
         else: print ".",
      print

def newgen(h,a,b,n):
   # Fuechse fressen Hasen oder sterben 
   # Fuchs ueberlebt, wenn mindestens ein Hase in der Naehe ist
   # ein Hase wird dann aber gefressen
   for i in range(1,n+1):
      for j in range(1,n+1):
         if a[i][j]==-10:
            # berechne Anzahl der Hasen in der Umgebung des Fuchses
            hs=0
            for k in range(-1,2):
               for l in range(-1,2):
                  if a[i+k][j+l]==1:
		      hs=hs+1	 
            if hs < 1:
               # Fuchs verhungert
               a[i][j]=0
            else:
               # ein Hase in der Umgebung des Fuchses wird zufaellig ausgewaehlt und gefressen
               hs=1
               while hs==1:
                  x=random.randint(-1,1)
                  y=random.randint(-1,1)
                  if a[i+x][j+y]==1:
                     hs=0
                     a[i+x][j+y]=0
   # Fuchs kann sich nur vermehren, wenn mindestens ein zweiter Fuchs in der Naehe ist.
   # Dieser kann sich dann aber im aktuellen Zeitschritt nicht mit noch einem Fuchs vermehren.
   # Fuechse koennen sich nur vermehren, wenn in Nachbarschaft noch ein Feld ohne Fuchs
   # vorhanden ist.
   # Fuechse werden vor Vermehrung in separates Feld b kopiert, damit Fuechse nicht doppelt 
   # beruecksichtigt werden.
   copymat(a,b,n)
   for i in range(1,n+1):
      for j in range(1,n+1):
         if b[i][j]==-10:
            fplus=0
            for k in range(-1,2):
               for l in range(-1,2):
                  if (k!=0 or l!=0) and b[i+k][j+l]==-10:
                     if fplus == 0:
                        # zwei Fuechse gefunden, die sich vermehren
                        b[i+k][j+l]=0
                        b[i][j]=0
                     fplus=fplus + 1
                     # fplus enthaelt Info, ob noch Feld ohne Fuchs in Naehe vorhanden (fplus < 8)
            while fplus > 0 and fplus < 8 :
               # neuer Fuchs wird zufaellig in Naehe platziert; sollte Hase auf diesem Feld
               # gewesen sein, wird er gefressen (d.h. er verschwindet)
               x=random.randint(-1,1)
               y=random.randint(-1,1)
               if a[i+x][j+y]!=-10:
                  fplus=0
                  a[i+x][j+y]=-10
   # Fuechse bewegen sich zufaellig um bis zu 2 Feldern
   copymat(a,b,n)
   for i in range(1,n+1):
      for j in range(1,n+1):
         if b[i][j]==-10:
             x=random.randint(-2,2)
             y=random.randint(-2,2)
             ineu=max(1,i+x)
             jneu=max(1,j+y)
             ineu=min(n,i+x)
             jneu=min(n,j+y)
             if a[ineu][jneu]!=-10:
                a[ineu][jneu]=-10
                a[i][j]=0
   # Hasen vermehren sich, wenn zweiter Hase in der Naehe ist: in der Regel ein neuer Hasen
   # Ausnahmen: es sind bereits zu viele Nachbarfelder durch Hasen und Fuechse besetzt
   # Hasen vor Vermehrung in separates Feld kopieren, damit Hasen nicht doppelt 
   # beruecksichtigt werden
   copymat(a,b,n)
   clearmat(h,n)
   # a: alte Hasen
   # b: alte Hasen, die sich noch nicht vermehrt haben
   # h: alte und neue Hasen
   for i in range(0,n+2):
      for j in range(0,n+2):
         if b[i][j]==1:
            h[i][j]=1
   for i in range(1,n+1):
      for j in range(1,n+1):
         if b[i][j]==1:
            b[i][j]=0
            hplus=0 
            sumh=0 
            sumf=0
            for k in range(-1,2):
               for l in range(-1,2):
                  if (k!=0 or l!=0) and h[i+k][j+l]==1:
                     # 2 Hasen gefunden, die sich vermehren
                     b[i+k][j+l]=0
                     hplus=1
                     break
               if hplus==1:
                  break
            if hplus ==1:
               # bestimme Anzahl freier Nachbarzellen
               for k in range(-1,2):
                  for l in range(-1,2):
                     if h[i+k][j+l]==1:
                        sumh=sumh+1
                     elif a[i+k][j+l]==-10:
                        sumf=sumf+1
               hplus = min(hplus, 9-sumf-sumh)
            while hplus>0:
               # platziere neue Hasen zufaellig auf freie Zelle, die noch frei ist
               x=random.randint(-1,1)
               y=random.randint(-1,1)
               if h[i+x][j+y]==0 and a[i+x][j+y] !=-10:
                  hplus=hplus-1
                  h[i+x][j+y]=1
   # kopiere aktuelle Info ueber Hasen in Matrix a
   for i in range(1,n+1):
      for j in range(1,n+1):
         if h[i][j]==1:
            a[i][j]=1
		       
   
#################
# HAUPTPROGRAMM #
#################

n=input("Anzahl von Zellen in jeder Dimension, z.B. 40:  ")
ngen=input("Anzahl von Generationen")
# definiere benoetigte Matrizen und vergroessere sie um einen aeusseren Randbereich,
# der die Programmierung vereinfacht, weil die Anzahl der Nachbarzellen am Rand
# dann in der Programmlogik nicht beruecksichtigt werden muss.
a=range(n+2)
b=range(n+2)
h=range(n+2)
for i in range(n+2):
   a[i]=range(n+2)
   b[i]=range(n+2)
   h[i]=range(n+2)
# initialisiere Matrizen a, b, h auf "0"
clearmat(a,n)
clearmat(b,n)
clearmat(h,n)
# belege Matrix a zufaellig zu etwa 50 % mit Hasen "1" und zu etwa 25 % mit Fuechsen "-10"
# speichere die Infos ueber die Hasen zusaetzlich in Matrix h
for i in range(1,n+1):
   for j in range(1,n+1):
      t=random.randint(0,4)
      if t==0 or t==1:
         h[i][j]=1
         a[i][j]=1
      elif t==2:
         a[i][j]=-10
      else:
         a[i][j]=0
# fuer unix- oder linux-systeme bitte den folgenden Befehl aktivieren:
# os.system("clear")
os.system("cls")
printmat(a,n)
history=range(3)
for i in range(3):
   history[i]=range(ngen+1)
# analysiere die aktuelle Population (d.h. bestimme die Anzahl von Hasen und Fuechsen)
analyze_population(a,n,history,ngen,0)
for it in range(ngen):
   print it
   # kopiere Matrix a auf b
   copymat(a,b,n)
   # berechne eine neue Generation von Hasen und Fuechsen
   newgen(h,a,b,n)
   # verzoegere Ausgabe um 0.5 Sekunden
   time.sleep(0.5)
   # loesche aktuellen Bildschirminhalt
   # fuer unix- oder linux-systeme bitte den folgenden Befehl aktivieren:
   # os.system("clear")
   os.system("cls")
   print
   # drucke den aktuellen Zustand aus
   printmat(a,n)
   # analysiere die aktuelle Population (d.h. bestimme die Anzahl von Hasen und Fuechsen)
   analyze_population(a,n,history,ngen,it+1)
# drucke die Historie der Hasen- und Fuchspopulation aus
print_history(history,ngen+1)
raw_input("zum Beenden des Programms Eingabetaste druecken")
